// ****************************************************************************
//
// Il modulo definisce le funzioni ausiliarie private necessarie alla
// implementazione delle funzioni di interfaccia al tipo PEG2.
//
// ****************************************************************************

// *************** IS INCIDENT ************************************************
// Controlla se l'edge denotato da E_ptr appartiene alla relazione VE totale 
// del vertice denotato da V_ptr senza usare la funzione di interfaccia 
// vertex_edge.

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_incident(v_type * V_ptr, e_type * E_ptr)
{
 e_type * e = V_ptr->incident_edge();

 if (e != NULL)
 {
   e_type * e1 = e;
   do
   { 
     if (e1 == E_ptr) 
       return true;
     e1 = e1->edge_incident_in_vertex(V_ptr);
   }
   while (e1 != e);   
 }
 return false;
}

// *************** IS VERTEX IN FACE ******************************************
// Controlla se la faccia denotata da F_ptr appartiene alla relazione VF totale 
// del vertice denotato da V_ptr senza usare la funzione di interfaccia 
// vertex_face.

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_vertex_in_face(v_type * V_ptr, f_type * F_ptr)
{
 if (V_ptr->is_isolated())
   return (F_ptr == V_ptr->containing_face());
   
 e_type * e = V_ptr->incident_edge();
 if (e != NULL)
 {
   e_type * e1 = e;
   do
   { 
     if (V_ptr == e1->source_vertex())
     {
       if (e1->right_face() == F_ptr) 
         return true;
     }
     else
     {
       if (e1->left_face() == F_ptr) 
         return true;
     }
     e1 = e1->edge_incident_in_vertex(V_ptr);
   }
   while (e1 != e);   
 }
 
 return false;
}

// ******************** IS ISOLATED FEATURE ***********************************
// Prende in input un vertice V_ptr appartenente ad una feature e ritorna true
// se la feature non ha contatti con un boundary.
// Usa il metodo delle liste affiancate per scorrere la feature.

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_isolated_feature(v_type * V_ptr)
{
  Vertex_List vl,twin;

  Edge_2 * e = V_ptr->incident_edge();
  if (e != NULL) 
  {
    vl.push_back(e->source_vertex());
    twin.push_back(NULL);
    Edge_2 * e0, * e1;
    Vertex_2 * v, * twin_v;
    while(!vl.empty())
    {
      v = vl.front();
      twin_v = twin.front();
      vl.pop_front();
      twin.pop_front();

      e0 = v->incident_edge();
      if (e0 != NULL)
      {
        e1 = e0;
        do
        { 
	  if (e1->left_face() == e1->right_face())
          {
            if (twin_v != e1->other_vertex(v))
            {
	      vl.push_back(e1->other_vertex(v));
              twin.push_back(v);
            }
          }
          else
            // se le due facce sono diverse sono su un boundary
            return false;
          e1 = e1->edge_incident_in_vertex(v);
        }
        while (e1 != e0);
      }
    } 
  }
  return true;
}

// ******************** NOT ON BOUNDARY ***************************************
// Prende in input un vertice V_ptr e ritorna true se non e' sul bordo di 
// nessuna regione

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
not_on_boundary(v_type * V_ptr)
{
  if (V_ptr->is_isolated())
    return true;
    
  e_type * e = V_ptr->incident_edge();
  if (e != NULL)
  {
    // lista degli spigoli associati alla relazione Vertex-Edge totale di V_ptr
    e_list_type el;
    vertex_edge(V_ptr, el);
    ei_type ei_end = el.end();
    for (ei_type ei = el.begin(); ei != ei_end; ei++)
      if (!(*ei)->is_in_feature())
        return false;
  }
  
  return true;
}

// ******************** IS ON BOUNDARY ****************************************
// Prende in input un vertice V_ptr e una regione F_ptr e ritorna true se e'
// sul bordo esterno della regione

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_on_boundary(v_type * V_ptr, f_type * F_ptr)
{
  if (V_ptr->is_isolated())
    return false;
    
  e_type * e = V_ptr->incident_edge();
  if (e != NULL)
  {
    // lista degli spigoli associati alla relazione Vertex-Edge totale di V_ptr
    e_list_type el;
    vertex_edge(V_ptr, el);
    ei_type ei_end = el.end();
    for (ei_type ei = el.begin(); ei != ei_end; ei++)
      if ((*ei)->is_bounding(F_ptr))
        return true;
  }
  
  return false;
}

// ******************** IS ON SOME BOUNDARY  **********************************
// Prende in input un vertice V_ptr e controlla se e' sul boundary di una 
// qualche regione.

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_on_some_boundary(v_type * V_ptr)
{
  if (V_ptr->is_isolated())
    return false;
    
  e_type * e = V_ptr->incident_edge();
  if (e != NULL)
  {
    // lista degli spigoli associati alla relazione Vertex-Edge totale di V_ptr
    e_list_type el;
    vertex_edge(V_ptr, el);
    ei_type ei_end = el.end();
    for (ei_type ei = el.begin(); ei != ei_end; ei++)  
      if (!(*ei)->is_in_feature())
        return true;
  }
  
  return false;
}

// ******************** GET NEXT BOUNDARY EDGE ********************************
// Assegna a E_ptr lo spigolo incidente in V_ptr incontrato ruotando in senso
// antiorario a partire dallo spigolo E0_ptr e che appartiene al boundary di
// qualche faccia.
// Si assume che E_ptr sia incidente in V_ptr.
// In caso lo spigolo non esista ritorna false ed E_ptr e' assegnato a NULL;

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
get_next_boundary_edge(v_type * V_ptr, e_type * E0_ptr, e_type *& E_ptr)
{
  if (E0_ptr->is_incident(V_ptr))
  {
    E_ptr = E0_ptr->edge_incident_in_vertex(V_ptr);
    while (E_ptr != E0_ptr)
    {
      if (!E_ptr->is_in_feature())
        return true;
      E_ptr = E_ptr->edge_incident_in_vertex(V_ptr);
    }
  }
  
  return false;
}

// ******************** EDGE IS REGION COMPONENT ******************************
// Prende in input uno spigolo E_ptr e una regione F_ptr ne ritorna se lo
// spigolo appartiene alla regione (e' sul contorno o in una feature).

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
edge_is_region_component(e_type * E_ptr, f_type * F_ptr)
{
  Edge_List_List ell;
  
  // Controllo se lo spigolo appartiene alla regione
  face_edge(F_ptr, ell); 	// spigoli dei bordi
  f_face_edge(F_ptr, ell);	// spigoli delle feature
  if (!ell.empty())
  {
    Edge_List el;
    Edge_List_Iterator eli_end = ell.end();
    for (Edge_List_Iterator eli = ell.begin(); eli != eli_end; eli++)
    {
      el = *eli;
      if (el.is_in_list(E_ptr))
        return true;
    }
  }  
  
  return false;
}

// ******************** GET EDGE POSITION *************************************
// Assegna ad ei la posizione di E_ptr nella lista el. Nel caso E_ptr non 
// appartenga alla lista ritorna false. Se E_ptr == NULL ritorna il past-end
// della lista.
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
get_edge_position(e_type * E_ptr, e_list_type & el,  ei_type & ei)
{
  ei_type ei_end = el.end();
  
  if (E_ptr == NULL)
  {
    ei = el.end();
    return true;
  }
  
  else  
  {  
    ei = el.begin();
    while ((ei != ei_end)&&(*ei != E_ptr))
      ei++;
    return (ei != ei_end);
  }
}

// ******************** UPDATE EDGE VERTEX  ***********************************
// Sia E1 il vertice rappresntato dall'iteratore ei1 nella lista el, ed E2
// quello rappresentato dall'iteratore ei2.
// Aggiorna la relazione edge-vertex dei vertici incidenti in V_ptr compresi
// in senso orario tra E1 (compreso) ed E2 (escluso), sostituendone
// il vertice V_ptr con il vertice V2_ptr. La lista el rappresenta i vertici
// incidenti in V_ptr. ek viene assegnato al vertice incontrato muovendosi
// in senso antiorario da E2, ovvero l'ultimo dei vertici modificati
// seguendo l'orientamento sopraesposto.
// Se ei1 indica il past-end della lista la funzione non esegue alcuna azione e
// ek viene assegnato a NULL.
// Se ei2 indica il past-end della lista tutti gli spigoli appartenenti alla
// lista vengono modificati. In tal caso ek viene aasegnato allo spigolo
// incontrato muovendosi in senso antiorario a partire da E1.
// Se necessario modifica le relazioni FIV delle facce coinvolte (che hanno
// nelle feature i vertici coinvolti nello spostamento)
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
update_edge_vertex(v_type * V_ptr, v_type * V2_ptr, ei_type ei1, ei_type ei2,
	           e_list_type & el, e_type *& ek)
{
  ek = NULL;
  
  ei_type ei, ei_end;
  ei = ei1;
  if (ei2 == el.end())
    ei_end = ei1;
  else
    ei_end = ei2;
  if (ei != el.end())
  {
    do
    {
      ek = * ei;     
      if (ek->source_vertex() == V_ptr)
        ek->assign_source_vertex(V2_ptr);
      if (ek->target_vertex() == V_ptr)
        ek->assign_target_vertex(V2_ptr);
      // se ek e' contenuta in una feature modifica la relazione FV della
      // faccia contenente ek (sostituisco V2_ptr a V_ptr nella lista dei
      // vertici interni alla faccia)
      if (ek->is_in_feature())
      {
        f_type * f = ek->left_face();
        if (f->is_f_vertex(V_ptr))
        {
          f->remove_f_vertex(V_ptr);
          f->add_f_vertex(V2_ptr);
        }
      }
      ei = el.clock(ei);
    }
    while(ei != ei_end);
  }
}

// ******************** FEATURE EDGES *****************************************
// Ritorna nella lista el tutti gli edge appartenenti alle feature di cui e e'
// il rappresentante (e compreso).
// E' stata ricavata direttamente dal corpo della f_face_edge

template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
feature_edges(e_type * e, e_list_type & el)
{
  v_list_type v_stack;
  e_list_type e_stack;
  v_type * v;

  f_type * f = e->left_face();
  
  if (e != NULL) 
  {
    v = e->source_vertex();
    e_type * e1 = e;
    do
    {
      if (e1->is_in_feature(f)) 
      {
        el.push_front(e1);
        e_stack.push_back(e1);
        v_stack.push_back(e1->other_vertex(v));
      }
      e1 = e1->edge_incident_in_vertex(v);
    }
    while (e1 != e);
      
    while(!e_stack.empty())
    {
      e = e_stack.front();
      v = v_stack.front();
      e_stack.pop_front();
      v_stack.pop_front();

      e1 = e->edge_incident_in_vertex(v);
      while (e1 != e)
      {
        if (e1->is_in_feature(f)) 
        {
          el.push_front(e1);
          e_stack.push_back(e1);
          v_stack.push_back(e1->other_vertex(v));
        }
        e1 = e1->edge_incident_in_vertex(v);
      }
    }
  }
}

// ******************** IS CONNECT TO FEATURE EDGES ***************************
// Ritorna se E_ptr e' gia' connesso a qualche edge appartenente ad una feature
// della faccia indicata dalla sua relazione EF.

template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_connect_to_feature_edges(e_type * E_ptr)
{
  e_list_type el;
  feature_edges(E_ptr, el);
  el.remove(E_ptr);
  if (el.empty())
    return false;

  return true;
}

// ******************** IS ON FEATURE EDGE ************************************
// ritorna se v appartiene a una feature di f e nel caso assegna ad ei la
// la posizione del rappresentante della feature nella lista associata alla
// relazione FIE
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_on_feature_edge(v_type * v, f_type * f, ei_type & ei)
{
  if (f->no_f_inside_edges())
    return false;
    
  ei_type ei_end = f->f_inside_edges_end();
  e_list_type el;
  for (ei = f->f_inside_edges_begin(); ei != ei_end; ei++)
  {
    // assegna ad el la lista degli spigoli che compongono la feature
    // rappresentata da e
    feature_edges(*ei, el);
    if (el.empty())
      return false;
      
    ei_type ei2_end = el.end();
    for (ei_type ei2 = el.begin(); ei2 != ei2_end; ei2++)
    {
      if ((*ei2)->is_incident(v))
        return true;
    }
    
    el.free();
  }

  return false;
}

// ******************** GET EDGE INCIDENT IN VERTEX ***************************
// controlla se nella lista el esiste uno spigolo incidente in v0 e lo assegna
// ad e. Se lo spigolo non esiste ritorna false
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
get_edge_incident_in_vertex(v_type * v0, e_list_type & el, e_type *& e)
{
  e = NULL;
  
  if (el.empty())
    return false;
    
  ei_type ei_end = el.end();
  for (ei_type ei = el.begin(); ei != ei_end; ei++)
  {
    e = * ei;      
    if (e->is_incident(v0))
      return true;
  }
    
  return false;
}

// ******************** GET EDGE INCIDENT IN VERTEX ***************************
// controlla se nella lista el esiste uno spigolo incidente in v0 ed assegna
// ad ei il suo iteratore. Se lo spigolo non esiste ritorna false
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
get_edge_incident_in_vertex(v_type * v0, e_list_type & el, ei_type & ei)
{
  ei_type ei_end = el.end();
  for (ei = el.begin(); ei != ei_end; ei++)
  {
    if ((*ei)->is_incident(v0))
      return true;
  }
    
  return false;
}

// ******************** PATH TO BOUNDARY **************************************
// memorizza in path1 il cammino dal vertice v (incluso) fino al boundary di
// f, assegna inoltre a e1_bound ed e2_bound gli spigoli sul boundary di f
// incidenti in v1_bound. Gli spigoli sono assegnati in modo che partendo da
// un segmento interno al f e incidente in v e1_bound sia incontrato ruotando
// in senso orario ed e2_bound ruotando in senso antiorario
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
path_to_boundary(v_type * v, f_type * f, e_list_type & path,
		 v_type *& v_bound, e_type *& e1_bound, e_type *& e2_bound)
{   
  v_bound = NULL;
  e1_bound = NULL;
  e2_bound = NULL;
    
  if (v->is_isolated())
    return false;
    
  e_type * ek;
  e_type * e = v->incident_edge();
    
  // nel caso v non sia sul boundary di f controllo se e' ad esso
  // connesso e nel caso ne memorizzo l'unico cammino possibile composto
  // esclusivamente da spigoli appartenenti a feature (se ve ne fosse piu' di
  // uno si formerebbe inevitabilmente una faccia e dunque vi sarebbe nel
  // percorso almeno uno spigolo non appartenente a feature)
  if (!is_on_boundary(v,f))
  {
    if (!e->is_in_feature())
      return false;
      
    if (e->source_edge() == e)
      v_bound = e->target_vertex();
    else if (e->target_edge() == e)
      v_bound = e->source_vertex();
    else
      return false;
    
    e_list_type el;
    feature_edges(e,el); // spigoli appartenenti a feature raggiungibili da e
    el.remove(e);
    
    ek = e;
    path.push_back(ek);
    while (!is_on_boundary(v_bound,f))
    {
      if (get_edge_incident_in_vertex(v_bound, el, ek))
      {
        v_bound = ek->other_vertex(v_bound);
        path.push_back(ek);
      }
      else
      {
        path.free();
      
        if (el.empty())
        {
          v_bound = NULL;
          return false;
        }
        else
        {
          el.remove(ek);
          ek = e;
          path.push_back(ek);
        }
      }
    }
  }
  else
    v_bound = v;
    
  ek = e;
  while ((e1_bound == NULL) || (e2_bound == NULL))
  {
    if (ek->is_bounding(f))
    {
      if (e1_bound == NULL)
        e1_bound = ek;
      else
        e2_bound = ek;
    }
    ek = ek->edge_incident_in_vertex(v_bound);
    if ((ek == e) && ((e1_bound == NULL) || (e2_bound == NULL)))
    {
      v_bound = NULL;
      e1_bound = NULL;
      e2_bound = NULL;
      path.free();
      return false;
    }
  }
    
  // riassegna e11_bound ed e2_bound in modo che a partire da un segmento
  // interno a f ed incidente in v e1_bound sia incontrato ruotando in senso
  // orario ed e2_bound ruotando in senso antiorario
  if (e1_bound->source_vertex() == v_bound)
  {
    if (e1_bound->right_face() == f)
    {
      ek = e1_bound;
      e1_bound = e2_bound;
      e2_bound = ek;
    }
  }
  else
  {
    if (e1_bound->left_face() == f)
    {
      ek = e1_bound;
      e1_bound = e2_bound;
      e2_bound = ek;
    }
  }
  
  return true;
}

// ******************** GET CONNECTED EDGES ***********************************
// Ritorna nella lista el tutti gli edge raggiungibili da e. E' stata
// ricavata direttamente dal corpo della features_edge()

template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
get_connected_edges(e_type * e, e_list_type & el)
{
  v_list_type v_stack;
  e_list_type e_stack;
  v_type * v;

  if (e != NULL) 
  {
    v = e->source_vertex();
    e_type * e1 = e;
    do
    {
      if (!e1->is_marked()) 
      {
        el.push_front(e1);
        e_stack.push_back(e1);
        v_stack.push_back(e1->other_vertex(v));
        e1->mark();
      }
      e1 = e1->edge_incident_in_vertex(v);
    }
    while (e1 != e);
      
    while(!e_stack.empty())
    {
      e = e_stack.front();
      v = v_stack.front();
      e_stack.pop_front();
      v_stack.pop_front();
      
      e1 = e->edge_incident_in_vertex(v);
      while (e1 != e)
      {
        if (!e1->is_marked()) 
        {
          el.push_front(e1);
          e_stack.push_back(e1);
          v_stack.push_back(e1->other_vertex(v));
          e1->mark();
        }
        e1 = e1->edge_incident_in_vertex(v);
      }
    }
    
    ei_type ei_end = el.end();
    for (ei_type ei = el.begin(); ei != ei_end; ei++)
      (*ei)->unmark();
  }
}

// ******************** CONNECTED VERTICES ************************************
// controlla se i due vertici passati in input sono connessi basandosi solo 
// sulla relazione VE dei vertici ed EV degli spigoli incontrati
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
connected_vertices(v_type * V1_ptr, v_type * V2_ptr)
{   
  if ((V1_ptr->is_isolated()) || (V2_ptr->is_isolated()))
    return false;
  
  // memorizzo gli spigoli raggiungibili da V1_ptr
  e_type * e = V1_ptr->incident_edge();
  e_list_type el;
  get_connected_edges(e,el);
  ei_type ei_end = el.end();
  for (ei_type ei = el.begin(); ei != ei_end; ei++)
    if ((*ei)->is_incident(V2_ptr))
      return true;
 
  return false;
}

// ******************** GET CONNECTED COMMON EDGES ****************************
// Ritorna nella lista el tutti gli edge raggiungibili da e e comuni alle due
// facce passate in input. E' stata ricavata direttamente dal corpo della
// get_connected_edges()

template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
get_connected_common_edges(e_type * e, f_type * f1, f_type * f2,
			   e_list_type & el)
{
  v_list_type v_stack;
  e_list_type e_stack;
  v_type * v;

  if (e != NULL) 
  {
    v = e->source_vertex();
    e_type * e1 = e;
    do
    {
      if ((!e1->is_marked())&&(e1->is_bounding(f1))&&(e1->is_bounding(f2)))
      {
        el.push_front(e1);
        e_stack.push_back(e1);
        v_stack.push_back(e1->other_vertex(v));
        e1->mark();
      }
      e1 = e1->edge_incident_in_vertex(v);
    }
    while (e1 != e);
      
    while(!e_stack.empty())
    {
      e = e_stack.front();
      v = v_stack.front();
      e_stack.pop_front();
      v_stack.pop_front();

      e1 = e->edge_incident_in_vertex(v);
      while (e1 != e)
      {
        if ((!e1->is_marked())&&(e1->is_bounding(f1))&&(e1->is_bounding(f2)))
        {
          el.push_front(e1);
          e_stack.push_back(e1);
          v_stack.push_back(e1->other_vertex(v));
          e1->mark();
        }
        e1 = e1->edge_incident_in_vertex(v);
      }
    }
    
    ei_type ei_end = el.end();
    for (ei_type ei = el.begin(); ei != ei_end; ei++)
      (*ei)->unmark();
  }
}

// ******************** GET EXTERNAL EDGES  ***********************************
// Prende in input una lista di spigoli el e controlla se sono connessi.
// Assegna ad E1_ptr ed E2_ptr i due spigoli estremi.
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
get_external_edges(e_list_type el, e_type *& E1_ptr, e_type *& E2_ptr)
{
  E1_ptr = NULL;
  E2_ptr = NULL;

  if (el.empty())
    return false;  
                   
  E1_ptr = E2_ptr = * el.begin();
  el.pop_front();
  
  v_type * vk = E1_ptr->source_vertex();
  ei_type ei_k;
  while (get_edge_incident_in_vertex(vk, el, ei_k))
  {
    E1_ptr = * ei_k;
    vk = E1_ptr->other_vertex(vk);
    el.erase(ei_k);
  }
    
  vk = E2_ptr->target_vertex();
  while (get_edge_incident_in_vertex(vk, el, ei_k))
  {
    E2_ptr = * ei_k;
    vk = E2_ptr->other_vertex(vk);
    el.erase(ei_k);
  }

  if (!el.empty())
  {
    E1_ptr = NULL;
    E2_ptr = NULL;
  
    return false;
  }
    
  return true;
}

// *************** GET BOUNDARY VERTICES **************************************
// Assegna a V_list la lista di puntatori ai vertici del boundary associato
// alla faccia F_ptr e rappresentato dallo spigolo E_ptr.
// N.B. : se V_list non e' vuota inserisce in testa alla lista.
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
get_boundary_vertices(f_type * F_ptr, e_type * E_ptr, v_list_type & V_list)
{
  if (F_ptr != NULL)
  {
    if (E_ptr != NULL)
    {
      v_type * v;
      if (F_ptr == E_ptr->left_face())
        v = E_ptr->target_vertex();
      else
        v = E_ptr->source_vertex();

      v_type * v1 = v;
      e_type * e1 = E_ptr;
      do
      { 
        if (e1->is_bounding(F_ptr)) 
        {
          if (!v1->is_marked())
          {
            V_list.push_back(v1);
            v1->mark();
          }
          v1 = e1->other_vertex(v1);
        }
        
        e1 = e1->edge_incident_in_vertex(v1);        
      }
      while (e1 != E_ptr);
      
      // smarca i vertici    
      vi_type vi_end = V_list.end();
      for (vi_type vi = V_list.begin(); vi != vi_end; vi++)
        (* vi)->unmark();
    }
  }
}

// *************** GET BOUNDARY EDGES *****************************************
// Assegna a E_list la lista di puntatori agli spigoli del boundary associato
// alla faccia F_ptr e rappresentato dallo spigolo E_ptr.
// N.B. : se E_list non e' vuota inserisce in testa alla lista.
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
get_boundary_edges(f_type * F_ptr, e_type * E_ptr, e_list_type & E_list)
{
  if (F_ptr != NULL)
  {
    if (E_ptr != NULL)
    {
      v_type * v;
      if (F_ptr == E_ptr->left_face())
        v = E_ptr->target_vertex();
      else
        v = E_ptr->source_vertex();

      v_type * v1 = v;
      e_type * e1 = E_ptr;
      do
      { 
        if (e1->is_bounding(F_ptr)) 
        {
          E_list.push_back(e1);
          v1 = e1->other_vertex(v1);
        }
        e1 = e1->edge_incident_in_vertex(v1);        
      }
      while (e1 != E_ptr);
    }
  }
}

// *************** GET BOUNDARY FACES *****************************************
// Assegna a F_list la lista di puntatori alle facce adiacenti al boundary
// associato alla faccia F_ptr e rappresentato dallo spigolo E_ptr.
// N.B. : se F_list non e' vuota inserisce in testa alla lista.
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
get_boundary_faces(f_type * F_ptr, e_type * E_ptr, f_list_type & F_list)
{
  if (F_ptr != NULL)
  {
    if (E_ptr != NULL)
    {
      v_type * v;
      if (F_ptr == E_ptr->left_face())
        v = E_ptr->target_vertex();
      else
        v = E_ptr->source_vertex();

      v_type * v1 = v;
      e_type * e1 = E_ptr;
      f_type * f;
      do
      { 
        if (e1->is_bounding(F_ptr)) 
        {
          f = e1->other_face(F_ptr);
          if (!f->is_marked())
          {
            f->mark();
            F_list.push_back(f);
          }
          v1 = e1->other_vertex(v1);
        }
        e1 = e1->edge_incident_in_vertex(v1);        
      }
      while (e1 != E_ptr);
    }
    
    // smarca le facce    
    fi_type fi_end = F_list.end();
    for (fi_type fi = F_list.begin(); fi != fi_end; fi++)
      (* fi)->unmark();
  }
}

// *************** GET NEXT FEATURE EDGE **************************************
// Prende in input lo spigolo E0_ptr che deve appartenere ad una feature e 
// assegna a Es_ptr e Et_ptr rispettivamente gli spigoli incidenti nel
// vertice source e nel vertice target di E0_ptr, successivi (in senso
// antiorario) a E0_ptr e appartenenti alla medesima feature. In caso di
// errore ritorna false e assegna Es_ptr e Et_ptr a NULL.
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
get_next_feature_edges(e_type * E0_ptr, e_type *& Es_ptr, e_type *& Et_ptr)
{
  Es_ptr = NULL;
  Et_ptr = NULL;
  
  if (!E0_ptr->is_in_feature())
    return false;
    
  f_type * f = E0_ptr->left_face();
  
  v_type * vk = E0_ptr->source_vertex();
  Es_ptr = E0_ptr;
  do
    Es_ptr = Es_ptr->edge_incident_in_vertex(vk);
  while ((Es_ptr != E0_ptr) && (!Es_ptr->is_in_feature(f)));
  
  vk = E0_ptr->target_vertex();
  Et_ptr = E0_ptr;
  do
    Et_ptr = Et_ptr->edge_incident_in_vertex(vk);
  while ((Et_ptr != E0_ptr) && (!Et_ptr->is_in_feature(f)));
  
  return true;
}

// *************** UPDATE FEATURE REL *****************************************
// Si suppone che E_ptr appartenesse ad una feature di F_ptr e sia stato da
// essa eliminato aggiornando le relazioni EE degli spigoli in esso incidenti.
// Si suppone inoltre che E0_ptr ed E1_ptr fossero rispettivamente gli spigoli
// incidenti nel vertice source e nel vertice target di E_ptr, ad esso
// successivi (in senso antiorario) e appartenenti alla medesima feature.
// Aggiorna le relazioni FIE di F_ptr.
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
update_feature_rel(f_type * F_ptr, e_type * E_ptr, e_type * E0_ptr, 
		   e_type * E1_ptr)
{
  e_type * ek;
  ei_type ei_end = F_ptr->f_inside_edges_end();
  for (ei_type ei = F_ptr->f_inside_edges_begin(); ei != ei_end; ei++)
  {
    ek = * ei;
    if (ek == E_ptr) // E_ptr e' il rappresentante di una feature
    {
      // si spezza la catena delle componenti connesse
      F_ptr->remove_f_inside_edge(ei);
      if (E0_ptr != E_ptr)
        F_ptr->add_f_inside_edge(E0_ptr);
      if (E1_ptr != E_ptr)
        F_ptr->add_f_inside_edge(E1_ptr);
      break;
    }
    else  // E_ptr non e' il rappresentante della feature
    {
      // se nella feature rappresentata da ek e' contenuto E0_ptr [E1_ptr], da
      // E0_ptr [E1_ptr] si arrivava a E_ptr (ovvero E_ptr apparteneva alla 
      // feature ma non ne era il rappresentante). Poiche' si e' eliminato
      // E_ptr dalla feature, E1_ptr [E0_ptr] non e' piu' raggiungibile,
      // bisogna aggiungerlo come rappresentante del ramo di catena spezzatosi.
      e_list_type el;
      feature_edges(ek,el);
      if (el.is_in_list(E0_ptr))
      {
        if (E1_ptr != E_ptr)
        {
          F_ptr->add_f_inside_edge(E1_ptr);
          break;
        }
      }
      if (el.is_in_list(E1_ptr))
      {
        if (E0_ptr != E_ptr)
        {
          F_ptr->add_f_inside_edge(E0_ptr);
          break;
        }
      }
    }
  }
}

// *************** IS INSIDE FACE ********************************************
// Controlla se F_ptr rappresenta un buco di PF_ptr
template <class VInfo, class EInfo, class FInfo>
bool PEG2<VInfo,EInfo,FInfo>::
is_inside_face(f_type * F_ptr, f_type * PF_ptr)
{
  e_list_type el;
  get_boundary_edges(F_ptr, F_ptr->bounding_edge(), el);
  
  ei_type ei_end = el.end();
  for (ei_type ei = el.begin(); ei != ei_end; ei++)
    if (!(* ei)->is_bounding(PF_ptr))
      return false;
  
  return true;
}

// ******************** MARK FEATURE EDGES ************************************
// Ritorna nella lista el tutti gli edge appartenenti alle feature di cui e e'
// il rappresentante (e compreso) dopo averli marcati.
// E' stata ricavata direttamente dalla feature_edges().
// N.B. : Se el non e' vuota inserisce in testa alla lista.

template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
mark_feature_edges(e_type * e, e_list_type & el)
{
  v_list_type v_stack;
  e_list_type e_stack;
  v_type * v;

  f_type * f = e->left_face();
  
  if (e != NULL) 
  {
    v = e->source_vertex();
    e_type * e1 = e;
    do
    {
      if (e1->is_in_feature(f)) 
      {
        el.push_front(e1);
        e_stack.push_back(e1);
        e1->mark();
        v_stack.push_back(e1->other_vertex(v));
      }
      e1 = e1->edge_incident_in_vertex(v);
    }
    while (e1 != e);
      
    while(!e_stack.empty())
    {
      e = e_stack.front();
      v = v_stack.front();
      e_stack.pop_front();
      v_stack.pop_front();

      e1 = e->edge_incident_in_vertex(v);
      while (e1 != e)
      {
        if (e1->is_in_feature(f)) 
        {
          el.push_front(e1);
          e_stack.push_back(e1);
          e1->mark();
          v_stack.push_back(e1->other_vertex(v));
        }
        e1 = e1->edge_incident_in_vertex(v);
      }
    }
  }
}

// *************** MARK BOUNDARY EDGES ****************************************
// Assegna a E_list la lista di puntatori agli spigoli del boundary associato
// alla faccia F_ptr e rappresentato dallo spigolo E_ptr dop averli marcati.
// E' stata ricavata direttamente dalla get_boundary_edges().
// N.B. : se E_list non e' vuota inserisce in testa alla lista.
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
mark_boundary_edges(f_type * F_ptr, e_type * E_ptr, e_list_type & E_list)
{
  if (F_ptr != NULL)
  {
    if (E_ptr != NULL)
    {
      v_type * v;
      if (F_ptr == E_ptr->left_face())
        v = E_ptr->target_vertex();
      else
        v = E_ptr->source_vertex();

      v_type * v1 = v;
      e_type * e1 = E_ptr;
      do
      { 
        if (e1->is_bounding(F_ptr)) 
        {
          E_list.push_back(e1);
          e1->mark();
          v1 = e1->other_vertex(v1);
        }
        e1 = e1->edge_incident_in_vertex(v1);        
      }
      while (e1 != E_ptr);
    }
  }
}

// *************** REFINE_FACE_RELATIONS ******************************************
// Controlla e corregge le relazioni associate alla faccia F_ptr eliminando
// eventuali entita' duplicate.
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::
refine_face_relations(f_type * F_ptr)
{
  // relazione F-Face-Vertex
  vi_type vi_end = F_ptr->f_vertices_end();
  for (vi_type vi = F_ptr->f_vertices_begin(); vi != vi_end; vi++)
  {
    if ((* vi)->is_marked())
      F_ptr->remove_f_vertex(vi);
    else
      (* vi)->mark();
  }
  vi_end = F_ptr->f_vertices_end();
  for (vi = F_ptr->f_vertices_begin(); vi != vi_end; vi++)
    (* vi)->unmark();
    
  // relazione F-Face-Edge
  e_list_type el;
  ei_type ei_end = F_ptr->f_inside_edges_end();
  for (ei_type ei = F_ptr->f_inside_edges_begin(); ei != ei_end; ei++)
  {
    if ((* ei)->is_marked())
      F_ptr->remove_f_inside_edge(ei);
    else
      mark_feature_edges(* ei, el);
  }
  ei_end = el.end();
  for (ei = el.begin(); ei != ei_end; ei++)
    (* ei)->unmark();
  
  // relazione Face-Edge
  el.free();
  ei_end = F_ptr->inside_edges_end();
  for (ei = F_ptr->inside_edges_begin(); ei != ei_end; ei++)
  {
    if ((* ei)->is_marked())
      F_ptr->remove_inside_edge(ei);
    else
      mark_boundary_edges(F_ptr, * ei, el);
  }
  ei_end = el.end();
  for (ei = el.begin(); ei != ei_end; ei++)
    (* ei)->unmark();
}
