// ****************************************************************************
//
// Il modulo definisce i tipi estesi vertex_2< ... >, edge_2< ... > e
// face_2< ... > istanziando i tipi parametrici base vertex< ... >, edge< ... >
// e face< ... > (definiti nel file dcel.h), e il tipo PEG2< ... > a
// rappresentare una suddivisione piana generalizzata. VInfo, EInfo e FInfo
// rappresentano le informazioni ausiliarie che si desidera associare a
// vertex< ... >, edge< ... > e face< ... >. Per la correttezza di quanto
// definito tali informazioni devono essere o quelle definite nei tipi
// vertex_2_info< ... >, edge_2_info< ... >, face_2_info< ... > (vedi file
// "peg2_info.h") o in tipi da essi derivati.
//
// ----------------------
// DEFINIZIONE DEI TIPI :
// ----------------------
//
// In linea di principio, se non si ha bisogno di costruttori, il tipi estesi
// possono essere ricavati direttamente attraverso
//
//		vertex<VInfo,EInfo,FInfo>
//		edge<VInfo,EInfo,FInfo>
//		face<VInfo,EInfo,FInfo>
//
// Ad esempio, per ottenere il tipo esteso vertex_2< ... > e' sufficiente:
//
//		 typedef vertex<VInfo,EInfo,FInfo> vertex_2.
//
// La notevole importanza dei costruttori suggerisce pero' di creare una nuova 
// classe derivandola da quella base (e definire in essa i costruttori
// desiderati.
// Tornando all'esempio precedente il tipo vertex_2 puo' essere definito da:
//
//	template< ... >
//	struct vertex_2 : public vertex< ... >
//	{
//	  ...
//	  // costruttori
//	  ... 
//	}
//
// Mentre nel primo caso il tipo vertex_2< ... > e il tipo vertex< ... >
// coincidono, nel secondo la corrispondenza e' oggettiva ma non formale
// in quando vertex_2< ... > e' considerato un nuovo tipo (derivato da
// vertex< ... >), benche', in effetti, non aggiunga alcun campo di
// informazione (aggiunge solo costruttori ed eventualmente funzioni proprie).
// Poiche' all'interno delle classi base vi sono riferimenti ad elementi di 
// tipo base, se interessa permettere il casting da tipo base a tipo esteso
// E' NECESSARIO CHE TRA I COSTRUTTORI DEFINITI NELLE CLASSI ESTESE SIANO 
// PRESENTI QUELLI CHE PERMETTONO IL CASTING.
// Se il passaggio da tipo base a tipo esteso interessa solo a livello di
// puntatori il costruttore non e' necessario.
// Se ad esempio si desidera permettere il casting da tipo vertex< ... > a
// tipo vertex_2< ... > si dovra' inserire nella classe vertex_2< ... > un
// costruttore del tipo:
//
//  	vertex_2(vertex<...> const &) { ... };
//
// Se il passaggio da vertex< ... > a vertex_2< ... > interessa solo a livello
// di puntatori (come nel nostro caso) il costruttore non e' necessario.
// Vogliamo inoltre precisare che, poiche' le informazioni contenute nei tipi
// base sono tutte puntatori a strutture vertex< ... >, edge< ... > e
// face< ... > gli elementi da essi rappresentati non possono accedere
// alle eventuali funzioni definite in vertex_2< ... >, edge_2< ... > e
// face_2< ... >. Questo problema puo' essere risolto in due modi:
//
// 1) Definire tutte le funzioni ausiliarie nella strutture VInfo, EInfo ed
//    FInfo.
// 2) Assegnare i puntatori ad elemeti di tipo base a puntatori ad elementi di
//    tipo esteso prima di accedere alle funzioni.
//
// Supponiamo ad esempio che fun() sia  stata definita in edge_2< ... > e che
// Incident_Edge() sia una funzione propria di vertex< ... > che ritorni un 
// puntatore ad un struttura edge< ... > allora:
//
//   ...
//   vertex_2< ... > v;
//   edge_2< ... > * e;
//   ...
//   v.Incident_Edge()->fun(); // errore
//   ((edge_2< ... > *) v.Incident_Edge())->fun(); // ok
//   edge_2< ... > * e = (edge_2 *) v.Incident_Edge();
//   e->fun() // ok	
//
// Se fun() fosse stata definita in EdgeInfo non si sarebbe verificato alcun
// errore.
// Facciamo anche notare che il casting da puntatori a strutture vertex< ... >,
// edge< ... >, face< ... > a puntatori a strutture vertex_2< ... >,
// edge_2< ... >, face_2< ... >  deve essere esplicito (in quanto il C++ non
// supporta casting automatici da puntatori a strutture base a puntatori a
// strutture derivate).
// Per evitare ogni volta che si desidera accedere ad un dato puntatore di 
// dover esplicitare il casting opportuno sono state ridefinite in questo
// modulo le funzioni di accesso implementate nelle classi base (in pratica
// sono solo casting mascherati).
// Per ultimo facciamo notare che, al contrario delle funzioni di accesso, le 
// funzioni di assegnazione non presentano problemi poiche' il C++ supporta il
// casting implicito da puntatori a strutture derivate a puntatori a strutture
// base.
// Il fatto che le classi base da cui si ottengono le classi vertex_2< ... >,
// edge_2< ... >, face_2< ... > (come ad esempio la classe base
// CGAL_point_2< ... >) sono passate come parametri (cioe' sono accedute
// attraverso i parametri VInfo, EInfo ed FInfo) non permette la derivazione
// diretta dei costruttori definiti nelle classi base. Il problema e' comunque
// facilmente risolvibile utilizzando l'overloading dell'operatore " = ",
// ovvero supponiamo di voler definire il costruttore
// vertex_2(Type hx, Type hy) utilizzando il costruttore
// CGAL_Point_2(Type hx, Type hy), allora:
//
//	vertex_2(Type hx, Type hy) : CGAL_Point_2(hx,hy) // errore
//      {  
// 	  ... 
//	}
//
//	vertex_2(Type hx, Type hy)			// ok
//	{
//	  CGAL_Point_2< ... > p(hx,hy);
//	  (CGAL_Point_2 < ... >) * this = p;
//	  ...
//	}
//
// Cosi' facendo la parte di vertex_2< ... > derivata da CGAL_Point_2< ... >
// viene assegnata membro a membro attraverso la variabile locale p.
//
// -------------------------
// FUNZIONI DI INTERFACCIA :
// -------------------------
//
// Diamo di seguito l'elenco delle funzioni proprie associate alle classi
// definite. Allo scopo precisiamo che:
//
// - I tipi v_type, e_type, f_type sono ridefinizioni dei tipi
//   vertex_2< ... >, edge_2< ... > e face_2< ... >.
//     
// - I tipi v_list_type, e_list_type, f_list_type sono ridefinizioni dei tipi
//   list_2<* v_type>, list_2<* e_type>, list_2<* f_type>.
//
// - I tipi vi_type, ei_type, fi_type sono ridefinizioni dei tipi
//   v_list_type::iterator, e_list_type::iterator, f_list_type::iterator.
//
// ------------------------
// Classe vertex_2< ... > :
//-------------------------
//
// Oltre alle funzioni sottoelencate sono stati estesi a questa classe tutti
// i costruttori associati al tipo CGAL_Point_2< ... >.
//
// Funzioni di accesso "private" :
// -------------------------------
//
// - vi_type get_iterator() :
//   Ritorna l'iteratore che indica il vertice all'interno della corrispondente
//   lista associata alla suddivisione.
//
// Funzioni di assegnazione "private" :
// ------------------------------------
//
// - void assign_iterator_pointer(const vi_type * it) :
//   Assegna ad it il puntatore all'iteratore che indica il vertice all'interno
//   della corrispondente lista associata alla suddivisione .
//
//
// ----------------------
// Classe edge_2< ... > :
// ----------------------
//
// Funzioni di accesso "private" :
// -------------------------------
//
// - ei_type get_iterator() :
//   Ritorna l'iteratore che indica lo spigolo all'interno della corrispondente
//   lista associata alla suddivisione.
//
// Funzioni di assegnazione "private" :
// ------------------------------------
//
// - void assign_iterator_pointer(const ei_type * it) :
//   Assegna ad it il puntatore all'iteratore che indica lo spigolo all'interno
//   della corrispondente lista associata alla suddivisione .
//
// Funzioni di accesso "public" :
// ------------------------------
// 
// - e_type edge_incident_in_vertex(v_type * v)
//   Ritorna lo spigolo associato alla relazione Edge-Edge applicata in v.
//   Gestisce i cappi in senso antiorario ritornando sempre lo spigolo
//   associato al vertice di partenza (quello incontrato ruotando in senso
//   antiorario al di fuori della faccia racchiusa dal cappio).
//
// - bool is_closed() :
//   Ritorna "true" se lo spigolo rappresenta un cappio.
//
// - bool is_in_feature() :
//   Ritorna "true" se lo spigolo appartiene alla relazione F-Face-Edge di 
//   una qualche faccia.
//
// - bool is_in_feature(f_type * f) :
//   Ritorna "true" se lo spigolo appartiene alla relazione F-Face-Edge della
//   faccia f.
//  
// - bool is_bounding(f_type * f) :
//   Ritorna "true" se lo spigolo appartiene al contorno (bordo esterno o
//   ciclo inreno) della faccia f.
//  
// - bool is_isolated() :
//   Ritorna "true" se lo spigolo non ha spigoli incidenti.
//
// - f_type * counter_clock_face(v_type * v) :
//   Ritorna la faccia che si incontra ruotando in senso antiorario sul vertice
//   v. Gestisce i cappi orientati in senso antiorario ritornando sempre la
//   faccia esterna
//
// - f_type * clock_face(v_type * v) :
//   Ritorna la faccia che si incontra ruotando in senso orario sul vertice
//   v. Gestisce i cappi orientati in senso antiorario ritornando sempre la
//   faccia esterna
//
// Funzioni di assegnazione "public" :
// -----------------------------------
//
// - void assign_edge_incident_in_vertex(v_type * v, e_type * e) :
//   Assegna lo spigolo e come spigolo associato alla relazione Edge-Edge
//   applicata in v. Gestisce i cappi in senso antiorario assegnando sempre 
//   la relazione Edge-Edge associata al vertice di partenza.
//
//
// ----------------------
// Classe face_2< ... > :
// ----------------------
//
// Funzioni di accesso "private" :
// -------------------------------
//
// - fi_type get_iterator() :
//   Ritorna l'iteratore che indica la faccia all'interno della corrispondente
//   lista associata alla suddivisione.
//
// Funzioni di assegnazione "private" :
// ------------------------------------
//
// - void assign_iterator_pointer(const fi_type * it) :
//   Assegna ad it il puntatore all'iteratore che indica la faccia all'interno
//   della corrispondente lista associata alla suddivisione .
//
//
// -------------
// Classe PEG2 :
// -------------
// 
// Funzioni "private" :
// -------------------- 
//
// - void remove_entity(v_type *& v) :
//   Elimina il vertice v dalla suddivisione e dealloca la memoria ad esso
//   associata (compreso la copia dell'iteratore che lo rappresenta nella
//   corrispondente lista).
//
// - void remove_entity(e_type *& e) :
//   Elimina lo spigolo e dalla suddivisione e dealloca la memoria ad esso
//   associata (compreso la copia dell'iteratore che lo rappresenta nella
//   corrispondente lista).
//
// - void remove_entity(f_type *& f) :
//   Elimina la faccia f dalla suddivisione e dealloca la memoria ad essa
//   associata (compreso la copia dell'iteratore che la rappresenta nella
//   corrispondente lista).
//
// I commenti e i parametri associati alle seguenti funzioni sono inseriti 
// nel file di definizione euler2.h.
//
//  void create_point(v_type *&, v_type);
//  void create_line(v_type *, v_type *, e_type *&);
//  void create_face(e_type *, f_type *&);
//
// I commenti e i parametri associati alle seguenti funzioni sono inseriti 
// nel file di definizione geometry2.h.
//
// - bool is_incident(v_type *, e_type *)
// - bool is_vertex_in_face(v_type *, f_type *)
// - bool is_isolated_feature(v_type * V_ptr)
// - bool edge_is_region_component(e_type *, f_type *)
// - bool not_on_boundary(v_type *)
// - bool is_on_boundary(v_type *, f_type *)
// - bool is_on_some_boundary(v_type *)
// - bool get_edge_position(e_type *, e_list_type &, ei_type &)
// - void update_edge_vertex(v_type *, v_type *, ei_type, ei_type,
//			     e_list_type &, e_type *& )
// - void feature_edges(e_type *, e_list_type &)
// - bool is_connect_to_feature_edges(e_type *)
// - bool is_on_feature_edge(v_type *, f_type *, ei_type &)
// - bool get_edge_incident_in_vertex(v_type *, e_list_type &, e_type *&)
// - bool get_edge_incident_in_vertex(v_type *, e_list_type &, ei_type &)
// - bool path_to_boundary(v_type *, f_type *, e_list_type &, v_type *&,
//  		           e_type *&, e_type *&)
// - void get_connected_edges(e_type *, e_list_type &)
// - bool connected_vertices(v_type *, v_type *)
// - void get_connected_common_edges(e_type *, f_type *, f_type *,
//				     e_list_type &)
// - bool get_external_edges(e_list_type, e_type *&, e_type *&)
// - void get_boundary(f_type *, e_list_type &)
// - void get_boundary_vertices(f_type *, v_list_type &)
// - void get_boundary_faces(f_type *, f_list_type &)
// - bool get_next_feature_edges(e_type *, e_type *&, e_type *&)
// - void update_feature_rel(f_type *, e_type *, e_type *, e_type *)
// - bool is_inside_face(f_type *, f_type *)
// - void mark_feature_edges(e_type *, e_list_type &)
// - void mark_boundary_edges(f_type *, e_type *, e_list_type &)
// - void refine_face_relations(f_type *)
//  
// Funzioni "public" :
// -------------------
//
// - vi_type vertices_begin() :
//   Ritorna l'iteratore al primo elemento della lista associata ai vertici
//   della suddivisione.
//
// - vi_type vertices_end() :
//   Ritorna l'iteratore che rappresenta il past-end della lista associata
//   ai vertici della suddivisione.
//
// - ei_type edges_begin() :
//   Ritorna l'iteratore al primo elemento della lista associata agli spigoli
//   della suddivisione.
//
// - ei_type edges_end() :
//   Ritorna l'iteratore che rappresenta il past-end della lista associata
//   agli spigoli della suddivisione.
//
// - fi_type faces_begin() :
//   Ritorna l'iteratore al primo elemento della lista associata alle facce
//   della suddivisione.
//
// - fi_type faces_end()
//   Ritorna l'iteratore che rappresenta il past-end della lista associata
//   alle facce della suddivisione.
//  
// - bool no_vertices() :
//   Ritorna "true" se nella suddivisione non vi sono vertici.
//  
// - bool no_edges() :
//   Ritorna "true" se nella suddivisione non vi sono spigoli.  
//
// - bool no_faces() :
//   Ritorna "true" se nella suddivisione non vi sono facce.
//
// - void print_face(f_type * f) :
//   stampa su standard output le informazioni associate alla faccia f.
//  
// - void print_subdivision() :
//   stampa su standard output le informazioni associate alle entita'
//   (vertci, spigoli e facce) della suddivisione.
//
// I commenti e i parametri associati alle seguenti funzioni sono inseriti 
// nel file di definizione relations2.h.
//
// - f_type * infinity();
// - bool is_infinity_face(f_type *)
// - void vertex_vertex(v_type *, v_list_type &)
// - void vertex_edge(v_type *, e_list_type &)
// - void vertex_face(v_type *, f_list_type &)
// - void edge_vertex(e_type *, v_type *&, v_type *&)
// - void edge_edge(e_type *, e_type *&, e_type *&, e_type *&, e_type *&)
// - void edge_face(e_type *, f_type *&, f_type *&)
// - void face_vertex(f_type *, list_2<v_list_type> &)
// - void face_edge(f_type *, list_2<e_list_type> &)
// - void face_face(f_type *, list_2<f_list_type> &)
// - void f_face_vertex(f_type *, v_list_type &)
// - void f_face_edge(f_type *, list_2<e_list_type> &)
//
// I commenti e i parametri associati alle seguenti funzioni sono inseriti 
// nel file di definizione euler2.h.
//
// - bool add_infinity_face(f_type *&)
// - bool add_point_feature(f_type *, v_type, v_type *&)
// - bool point_abstraction(v_type *&)
// - bool add_line_feature(v_type *, v_type *, e_type *, e_type *, e_type *&)
// - bool line_abstraction(e_type *&)
// - bool expand_point_to_line(v_type *, e_type *, e_type *, v_type,
//  			       v_type *&, v_type, v_type *&, e_type *&)
// - bool line_to_point_contraction(e_type *&, v_type, v_type *&)
// - bool expand_point_to_region(v_type *, e_type *, f_type *, e_type *&,
//  			         f_type *&)
// - bool region_to_point_contraction(f_type *&)
// - bool expand_line_to_region(e_type *, f_type *, f_type *&,  e_type *&)
// - bool region_to_line_contraction(f_type *&, e_type *&)
// - bool split_line(e_type *, v_type, v_type  *&, e_type *&)
// - bool line_merge(v_type *&, e_type *&)
// - bool split_region(f_type *,  v_type *, v_type *, e_type *&, f_type *&)
// - bool region_merge(e_type *&, f_type *&)
//
// ****************************************************************************

#ifndef __PEG2
#define __PEG2

#include <new.h>			// Tipi di CGAL.
#include <list_2.h>
#include <CGAL/Direction_2.h>
#include <CGAL/Polygon_2.h>
#include <CGAL/Circle_2.h>

#include <dcel.h>			// Tipo DCEL base.

template <class VInfo, class EInfo, class FInfo>
class edge_2;
template <class VInfo, class EInfo, class FInfo>
class face_2;
template <class VInfo, class EInfo, class FInfo>
class PEG2;


// ****************************************************************************
// 				VERTEX_2
// ****************************************************************************

template <class VInfo, class EInfo, class FInfo>
class vertex_2 : public vertex<VInfo,EInfo,FInfo>
{
  typedef VInfo::R::RT COORD_TYPE;	    // Tipo delle coordinate dei punti.
  typedef vertex_2<VInfo,EInfo,FInfo> v_type;
  typedef edge_2<VInfo,EInfo,FInfo> e_type;

  typedef list_2<v_type *>::iterator vi_type;
  typedef CGAL_Point_2<R> Point_2;

  friend class PEG2<VInfo,EInfo,FInfo>;

  void assign_iterator_pointer(const vi_type * it)
  { VInfo::assign_iterator_ptr((void *) it); }

  vi_type get_iterator()
  { return * ((vi_type *) VInfo::get_iterator_ptr()); }


  public:

  vertex_2() {}

  vertex_2(const CGAL_Origin & o)
  { (Point_2) * this = Point_2::operator=(Point_2(o)); }

  vertex_2(const Point_2 & p) 
  { (Point_2) * this = Point_2::operator=(p); }

  vertex_2(Point_2 & p) 
  { (Point_2) * this = Point_2::operator=(p); }

  vertex_2(const COORD_TYPE & hx, const COORD_TYPE & hy)
  { Point_2 p(hx,hy); (Point_2) * this = Point_2::operator=(p); }

  vertex_2(const COORD_TYPE & hx, const COORD_TYPE & hy,
	   const COORD_TYPE & hw)
  { Point_2 p(hx,hy,hw); (Point_2) * this = Point_2::operator=(p); }

  // Ridefinizione delle funzioni di accesso della classe vertex< ... >.
  e_type * incident_edge() 
  { return (e_type *) vertex<VInfo,EInfo,FInfo>::incident_edge(); }
};



// ****************************************************************************
// 				EDGE_2
// ****************************************************************************
template <class VInfo, class EInfo, class FInfo>
class edge_2 : public edge<VInfo,EInfo,FInfo>
{
  typedef vertex_2<VInfo,EInfo,FInfo> v_type;
  typedef edge_2<VInfo,EInfo,FInfo> e_type;
  typedef face_2<VInfo,EInfo,FInfo> f_type;
  typedef list_2<e_type *>::iterator_2 ei_type;

  friend class PEG2<VInfo,EInfo,FInfo>;

  void assign_iterator_pointer(const ei_type * it)
  { EInfo::assign_iterator_ptr((void *) it); }

  ei_type get_iterator()
  { return * ((ei_type *) EInfo::get_iterator_ptr()); }
  
  public:

  edge_2() {}

  v_type * source_vertex() 
  { return (v_type *) edge<VInfo,EInfo,FInfo>::source_vertex(); }
  
  v_type * target_vertex() 
  { return (v_type *) edge<VInfo,EInfo,FInfo>::target_vertex(); }
  
  e_type * source_edge()
  { return (e_type *) edge<VInfo,EInfo,FInfo>::source_edge(); }

  e_type * target_edge()
  { return (e_type *) edge<VInfo,EInfo,FInfo>::target_edge(); }
  
  f_type * left_face()
  { return (f_type *) edge<VInfo,EInfo,FInfo>::left_face(); }

  f_type * right_face()
  { return (f_type *) edge<VInfo,EInfo,FInfo>::right_face(); }

  v_type * other_vertex(v_type * v)
  { return (v_type *) edge<VInfo,EInfo,FInfo>::other_vertex(v); }

  e_type * next_clock_edge(f_type * f)
  { return (e_type *) edge<VInfo,EInfo,FInfo>::next_clock_edge(f); }

  f_type * other_face(f_type * f)
  { return (f_type *) edge<VInfo,EInfo,FInfo>::other_face(f); }

  v_type * common_vertex(e_type * e)
  { return (v_type *) edge<VInfo,EInfo,FInfo>::common_vertex(e); }

  // ---------------------------
  // Nuove funzioni di accesso :
  // ---------------------------
  
  e_type * edge_incident_in_vertex(v_type * v)
  {
    if (is_closed())
      return source_edge();
    else
     return ((e_type *) edge<VInfo,EInfo,FInfo>::edge_incident_in_vertex(v));
  }
  
  bool is_closed()
  { return (source_vertex() == target_vertex()); }
  
  bool is_in_feature()
  { return (left_face() == right_face()); }
  
  bool is_in_feature(f_type * f)
  { return ((left_face() == right_face()) && (left_face() == f)); }
  
  bool is_bounding(f_type * f)
  { return ((edge<VInfo,EInfo,FInfo>::is_bounding(f))&&(!is_in_feature(f))); }
    
  bool is_isolated()
  { return (source_edge() == this) && (target_edge() == this); }
  
  f_type * counter_clock_face(v_type * v)
  {
    f_type * f = NULL;
    if (source_vertex() == v)
      f = left_face();
    if (target_vertex() == v)
      f = right_face();
    return (f);
  }

  f_type * clock_face(v_type * v)
  {
    f_type * f = NULL;
    if (source_vertex() == v)
      f = right_face();
    if (target_vertex() == v)
      f = left_face();
    return (f);
  }
  
  // --------------------------------
  // Nuove funzioni di assegnazione :
  // --------------------------------
  
  void assign_edge_incident_in_vertex(v_type * v, e_type * e)
  {
     if (is_closed())
       assign_source_edge(e);
     else if (source_vertex() == v)
       assign_source_edge(e);
     else
       assign_target_edge(e);
   }
};



// ****************************************************************************
// 				FACE_2
// ****************************************************************************

template <class VInfo, class EInfo, class FInfo>
class face_2 : public face<VInfo,EInfo,FInfo>
{
  typedef edge_2<VInfo,EInfo,FInfo> e_type;
  typedef face_2<VInfo,EInfo,FInfo> f_type;

  typedef list_2<f_type *>::iterator_2 fi_type;
  typedef list_2<e_type *>::iterator_2 ei_type;

  friend class PEG2<VInfo,EInfo,FInfo>;

  void assign_iterator_pointer(const fi_type * it)
  { FInfo::assign_iterator_ptr((void *) it); }

  fi_type get_iterator()
  { return * ((fi_type *) FInfo::get_iterator_ptr()); } 


  public:

  face_2() {}

  // Ridefinizione delle funzioni di accesso della classe face< ...>.
  e_type * bounding_edge()
  { return (e_type *) face<VInfo,EInfo,FInfo>::bounding_edge(); }  
};


// ****************************************************************************
// 				      PEG2
// ****************************************************************************

template <class VInfo, class EInfo, class FInfo>
class PEG2
{
  typedef VInfo::R REP_TYPE;		// Tipo di rappresentazione dei punti
  typedef VInfo::R::RT COORD_TYPE;	// Tipo delle coordinate dei punti.

  
  typedef 				// Per l'utilizzo dei poligoni CGAL.
  CGAL_Polygon_2<REP_TYPE, list< CGAL_Point_2<REP_TYPE> > > POLYGON_2_TYPE;

  typedef vertex_2<VInfo,EInfo,FInfo> v_type;
  typedef edge_2<VInfo,EInfo,FInfo> e_type;
  typedef face_2<VInfo,EInfo,FInfo> f_type;

  typedef list_2<v_type *> v_list_type;
  typedef list_2<e_type *> e_list_type;
  typedef list_2<f_type *> f_list_type;

  typedef list_2<v_type *>::iterator vi_type;
  typedef list_2<e_type *>::iterator ei_type;
  typedef list_2<f_type *>::iterator fi_type;

  list_2<v_type *> vertex_list;
  list_2<e_type *> edge_list;
  list_2<f_type *> face_list;
  f_type * infinity_face;

  void remove_entity(v_type *& v)
  { 
    vertex_list.erase_ptr(v->get_iterator());
    v->free_iterator_ptr();
    v = NULL;
  } 

  void remove_entity(e_type *& e)
  { 
    edge_list.erase_ptr(e->get_iterator());
    e->free_iterator_ptr();
    e = NULL;
  } 

  void remove_entity(f_type *& f)
  { 
    face_list.erase_ptr(f->get_iterator());
    f->free_iterator_ptr();
    f = NULL;
  } 
  
  // funzioni definite in euler2.h
  void create_point(v_type *&, v_type);
  void create_line(v_type *, v_type *, e_type *&);
  void create_face(e_type *, f_type *&);

  // funzioni definite in geometry2.h
  bool is_incident(v_type *, e_type *);
  bool is_vertex_in_face(v_type *, f_type *);
  bool is_isolated_feature(v_type * V_ptr);
  bool edge_is_region_component(e_type *, f_type *);
  bool not_on_boundary(v_type *);
  bool is_on_boundary(v_type *, f_type *);
  bool is_on_some_boundary(v_type *);
  bool get_next_boundary_edge(v_type *, e_type *, e_type *&);
  bool get_edge_position(e_type *, e_list_type &, ei_type &);
  void update_edge_vertex(v_type *, v_type *, ei_type, ei_type, e_list_type &,
  			  e_type *& );
  void feature_edges(e_type *, e_list_type &);
  bool is_connect_to_feature_edges(e_type *);
  bool is_on_feature_edge(v_type *, f_type *, ei_type &);
  bool get_edge_incident_in_vertex(v_type *, e_list_type &, e_type *&);
  bool get_edge_incident_in_vertex(v_type *, e_list_type &, ei_type &);
  bool path_to_boundary(v_type *, f_type *, e_list_type &, v_type *&,
  		        e_type *&, e_type *&);  
  void get_connected_edges(e_type *, e_list_type &);
  bool connected_vertices(v_type *, v_type *);
  void get_connected_common_edges(e_type *, f_type *, f_type *, e_list_type &);
  bool get_external_edges(e_list_type, e_type *&, e_type *&);
  void get_boundary_vertices(f_type *, e_type *, v_list_type &);
  void get_boundary_edges(f_type *, e_type *, e_list_type &);
  void get_boundary_faces(f_type *, e_type *, f_list_type &);
  bool get_next_feature_edges(e_type *, e_type *&, e_type *&);
  void update_feature_rel(f_type *, e_type *, e_type *, e_type *);
  bool is_inside_face(f_type *, f_type *);
  void mark_feature_edges(e_type *, e_list_type &);
  void mark_boundary_edges(f_type *, e_type *, e_list_type &);
  void refine_face_relations(f_type *);
  
  public:
 
  PEG2()
  { infinity_face = NULL; }

  // N.B. : la faccia infinita e' deallocata da face_list.free_ptr().
  ~PEG2()
  { vertex_list.free_ptr(); edge_list.free_ptr(); face_list.free_ptr(); }

  vi_type vertices_begin()
  { return vertex_list.begin(); }

  vi_type vertices_end()
  { return vertex_list.end(); }

  ei_type edges_begin()
  { return edge_list.begin(); }

  ei_type edges_end()
  { return edge_list.end(); }

  fi_type faces_begin()
  { return face_list.begin(); }

  fi_type faces_end()
  { return face_list.end(); }
  
  bool no_vertices()
  { return vertex_list.empty(); }
  
  bool no_edges()
  { return edge_list.empty(); }
  
  bool no_faces()
  { return face_list.empty(); }

  void print_face(f_type * f);
  void print_subdivision();
  
  // funzioni definite in relations2.h
  f_type * infinity();
  bool is_infinity_face(f_type *);
  void vertex_vertex(v_type *, v_list_type &);
  void vertex_edge(v_type *, e_list_type &);
  void vertex_face(v_type *, f_list_type &);
  void edge_vertex(e_type *, v_type *&, v_type *&);
  void edge_edge(e_type *, e_type *&, e_type *&, e_type *&, e_type *&);
  void edge_face(e_type *, f_type *&, f_type *&);
  void face_vertex(f_type *, list_2<v_list_type> &);
  void face_edge(f_type *, list_2<e_list_type> &);
  void face_face(f_type *, list_2<f_list_type> &);
  void f_face_vertex(f_type *, v_list_type &);
  void f_face_edge(f_type *, list_2<e_list_type> &);

  // funzioni definite in euler2.h
  bool add_infinity_face(f_type *&);
  bool add_point_feature(f_type *, v_type, v_type *&);
  bool point_abstraction(v_type *&);
  bool add_line_feature(v_type *, v_type *, e_type *, e_type *, e_type *&);
  bool line_abstraction(e_type *&);
  bool expand_point_to_line(v_type *, e_type *, e_type *, v_type,
  			    v_type *&, v_type, v_type *&, e_type *&);
  bool line_to_point_contraction(e_type *&, v_type, v_type *&);
  bool expand_point_to_region(v_type *, e_type *, f_type *, e_type *&,
  			      f_type *&);
  bool region_to_point_contraction(f_type *&);
  bool expand_line_to_region(e_type *, f_type *&,  e_type *&);
  bool region_to_line_contraction(f_type *&, e_type *&);
  bool split_line(e_type *, v_type, v_type  *&, e_type *&);
  bool line_merge(v_type *&, e_type *&);
  bool split_region(f_type *,  v_type *, v_type *, e_type *&, f_type *&);
  bool region_merge(e_type *&, f_type *&);

  // esportazione dei tipi.
  typedef v_type Vertex_2;
  typedef e_type Edge_2;
  typedef f_type Face_2;
 
  typedef v_list_type Vertex_List;
  typedef e_list_type Edge_List;
  typedef f_list_type Face_List;

  typedef list_2<Vertex_List> Vertex_List_List;
  typedef list_2<Edge_List> Edge_List_List;
  typedef list_2<Face_List> Face_List_List;
  
  typedef Vertex_List::iterator Vertex_Iterator;
  typedef Edge_List::iterator Edge_Iterator;
  typedef Face_List::iterator Face_Iterator;

  typedef Vertex_List_List::iterator Vertex_List_Iterator;
  typedef Edge_List_List::iterator Edge_List_Iterator;
  typedef Face_List_List::iterator Face_List_Iterator;
};

#include <relations2.h>  // Funzioni interne di PEG2 (public).
#include <geometry2.h>   // Funzioni interne di PEG2 (public).
#include <euler2.h>	// Funzioni interne di PEG2 (public).


//************************* OUT OF STORE ************************************** 
// Handler per l'operatore di allocazione di memoria new()

void out_of_store()
{
  cerr << "Errore : Memoria Insufficiente" << endl;
  exit(1);
}

// ******************** PRINT FACE ********************************************
// Stampa su standard output le informazioni associate alla faccia f
template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::print_face(f_type * f)
{
  if (is_infinity_face(f))
    cout << "INFINITY ";
    
  cout << "FACE: [" << f << "] (" << *f->get_iterator() << ")" << endl;
  // spigoli dei bordi
  cout << "\tBORDER:" << endl;
  if ((f->no_inside_edges()) && (f->bounding_edge() == NULL))
    cout << "\t\tNO BORDER" << endl;
  else
  {
    Edge_List_List ell;
    face_edge(f, ell);
    Edge_List el;
    Edge_List_Iterator eli_end = ell.end();
    for (Edge_List_Iterator eli = ell.begin(); eli != eli_end; eli++)
    {
      cout << "\t\t\t==========" << endl;
      el = *eli;
      Edge_Iterator ei_end = el.end();
      for (Edge_Iterator ei = el.begin(); ei != ei_end; ei++)
      {
        Edge_2 *e = *ei;
        cout << "\t\tEdge -> [" << e << "]" << endl;
        cout << "\t\tEV -> " << *e << endl;
        cout << "\t\tEE -> "<< "(" << e->source_edge() << ",";
        cout << e->target_edge() << ")" << endl;
        cout << "\t\tEF -> "<< "(" << e->left_face() << ",";
        cout << e->right_face() << ")" << endl;
      }
      cout << "\t\t\t==========" << endl;
    }
  }
  // vertici e spigoli delle features
  cout << "\tFEATURE:" << endl;
  if ((f->no_f_inside_edges()) && (f->no_f_vertices()))
    cout << "\t\tNO FEATURE" << endl;
  else
  {
    cout << "\t\tVERTICES:" << endl;
    if (f->no_f_vertices())
      cout << "\t\t\tNO VERTICES" << endl;
    else
    {
      Vertex_Iterator vi_end = f->f_vertices_end();
      for (Vertex_Iterator vi = f->f_vertices_begin(); vi != vi_end; vi++)
      {
        Vertex_2 * v = * vi;;
        cout << "\t\t\t[" << v << "]";
        if (v != NULL)
        {
          cout << ": "<< * v << endl;
          cout << "\t\t\tVE -> " << v->incident_edge() << endl;
          cout << "\t\t\tVF -> " << v->containing_face();
        }
        cout << endl;
      }
    }
    cout << "\t\tEDGES:" << endl;
    if (f->no_f_inside_edges())
      cout << "\t\t\tNO EDGES" << endl;
    else
    {
      Edge_List_List ell;
      f_face_edge(f, ell);
      Edge_List el;
       Edge_List_Iterator eli_end = ell.end();
      
      for (Edge_List_Iterator eli = ell.begin(); eli != eli_end; eli++)
      {
        cout << "\t\t\t==========" << endl;
        el = *eli;
        Edge_Iterator ei_end = el.end();
        for (Edge_Iterator ei = el.begin(); ei != ei_end; ei++)
        {
          Edge_2 *e = *ei;
          cout << "\t\t\tEdge -> [" << e << "]" << endl;
          cout << "\t\t\tEV -> " << *e << endl;
          cout << "\t\t\tEE -> "<< "(" << e->source_edge() << ",";
          cout << e->target_edge() << ")" << endl;
          cout << "\t\t\tEF -> "<< "(" << e->left_face() << ",";
          cout << e->right_face() << ")" << endl;
        }
        cout << "\t\t\t==========" << endl;
      }
    }
  }
}

// ******************** PRINT SUBDIVISION *************************************
// Stampa su standard output le informazioni associate alle entita' della 
// (vertici, spigoli e facce) della suddivisione.

template <class VInfo, class EInfo, class FInfo>
void PEG2<VInfo,EInfo,FInfo>::print_subdivision()
{
  cout << "****************************************" << endl;
  cout << "********* SUBDIVISION (begin) **********" << endl << endl;
  cout << "----------- VERTICES (begin) -----------" << endl;
  if (no_vertices())
    cout << "NO VERTICES" << endl;
  else
  {
    Vertex_Iterator vi_end = vertices_end();
    for (Vertex_Iterator vi = vertices_begin(); vi != vi_end; vi++)
    {
      Vertex_2 * v = * vi;;
      cout << "[" << v << "]";
      if (v != NULL)
      {
        cout << ": "<< * v << endl;
        cout << "\tVE -> " << v->incident_edge() << endl;
        cout << "\tVF -> " << v->containing_face();
      }
      cout << endl;
    }
  }
  cout << "------------ VERTICES (end) ------------" << endl << endl;
  cout << "------------- EDGES (begin) ------------" << endl;
  if (no_edges())
    cout << "NO EDGES" << endl;
  else
  {
    Edge_Iterator ei_end = edges_end();
    for (Edge_Iterator ei = edges_begin(); ei != ei_end; ei++)
    {
        Edge_2 *e = *ei;
        cout << "Edge -> [" << e << "]" << endl;
        cout << "\tEV -> " << *e << endl;
        cout << "\tEE -> "<< "(" << e->source_edge() << ",";
        cout << e->target_edge() << ")" << endl;
        cout << "\tEF -> "<< "(" << e->left_face() << ",";
        cout << e->right_face() << ")" << endl;
    }
    cout << endl;
  }
  cout << "------------- EDGES (end) --------------" << endl << endl;
  cout << "------------- FACES (begin) ------------" << endl;
  if (no_faces())
    cout << "NO FACES" << endl;
  else
  {
    Face_Iterator fi_end = faces_end();
    for (Face_Iterator fi = faces_begin(); fi != fi_end; fi++)
      print_face(*fi);
  }
  cout << "-------------- FACES (end) -------------" << endl << endl;
  cout << "********** SUBDIVISION (end) ***********" << endl;
  cout << "****************************************" << endl;
}

#endif
