// ****************************************************************************
//
// Il modulo definisce i tipi per le informazioni ausiliarie associate alle
// strutture vertex_2< ... >, edge_2< ... > e face_2< ... > (vedi file
// "peg2.h").
//
// L'UTILIZZO DI QUESTO MODULO PREVEDE L'INCLUSIONE PRECEDENTE DA DA PARTE 
// DELL'UTENTE DEI SEGUENTI MODULI CGAL:	 
//								  
// - Cartesian.h e/o Homogeneous.h
//
// ESEMPIO:
// #include <CGAL/Cartesian.h>
// #include <subdivision_2_info.h>
//
// --------------------------------
// DEFINIZIONE DEI TIPI AUSILIARI :
// --------------------------------
//
// I tipi definiti sono :
//
// Struttura vertex_2_info<RepType,Ext>
// {
//	 struttura Point_2 di CGAL
//           +
//	 struttura Ext
//	     +
//	 campo a tipo bool
//	     +
// 	 campo puntatore a tipo void
//	     +
//	 campo puntatore a tipo face_2
// }
//	 		     
// Struttura edge_2_info<Ext> 
// {
//	 struttura Ext
//	     +
//	 campo a tipo bool
//	     +
//	 campo puntatore a tipo void
// }
//
// Struttura face_2_info<RepType,Ext>
// {
// 	 struttura Ext
//	     +
//	 campo a tipo bool
//	     +
//	 campo puntatore a tipo void
//	     +
//	 campo a tipo lista di puntatori a tipo vertex_2< ... >
//	     +
//	 campo a tipo lista di puntatori a tipo edge_2< ... >
//           +
//       campo a tipo lista di puntatori a tipo edge_2< ... >
// }
//
// La struttura vertex_2_info<RepType,Ext> rappresenta:
// - Le coordinate cartesiane del punto in rappresentazione cartesiana od
//   omogenea a seconda che l'istanza del parametro RepType sia
//   CGAL_Cartesian<RepType> o CGAL_Homogeneous<RepType>.
// - La faccia associata alla relazione F-Vertex-Face.
//
// La struttura face_2_info<RepType,Ext> rappresenta:
// - Un campo di marcatura per la faccia.
// - La lista dei vertici associata alla relazione F-Face-Vertex.
// - La lista degli spigoli associati alla relazione F-Face-Edge.
// - la lista di spigoli ausiliaria associata alla relazione Face-Edge.
//
// In ogni struttura e' presente inoltre un campo di marcatura e il puntatore
// di tipo "void" iterator_ptr. Attraverso di esso si puo' implementare
// l'accesso diretto all'interno di liste contenitore (ovvero si puo' accedere
// direttamente all'iteratore che identifica l'elemento all'interno della
// lista in questione). Allo scopo e' necessario implementare le seguenti fasi:
//
// - Definire una funzione di assegnazione ed una funzione di accesso al
//   campo iterator_ptr che realizzino gli opportuni casting.
//   La funzione di assegnazione effettua il casting da tipo iteratore a tipo
//   void, la funzione di accesso da tipo void a tipo iteratore.
//   Tali funzioni sono state implementate nelle classi vertex_2< ... >, 
//   edge_2< ... > e face_2< ... >.
// - Una volta inserito l'elemento in una lista, assegnare (attraverso la
//   funzione citata sopra) il campo iterator_ptr al puntatore l'iteratore.
//
// Il parametro Ext rappresenta le eventuali informazioni aggiuntive richieste
// dall'utente.
// Nel caso si voglia aggiungere delle informazioni e' quindi sufficiente
// progettare nuove classi che rappresentino le informazioni desiderate e
// istanziare il parametro Ext su tali classi.
//
// Il tipo lista da noi utilizzato (definito in list_2.h) deriva direttamente
// dal tipo lista di CGAl ( definito in list.h) che importa direttamente il 
// tipo di dato dalla libreria standard STL.
// Poiche' le funzioni di accesso ritornano iteratori (e non reference ad
// iteratori) non e' possibile assegnare ad iterator_ptr l'indirizzo del
// valore di ritorno. (in quanto esso rappresenta l'indirizzo di una variabile
// temporanea).
// Inoltre, poiche' la struttura iteratore in list.h e' definita "private",
// non e' possibile definire in list_2.h funzioni che ritornino "reference" 
// a tale struttura.
// Il problema e' stato risolto copiando l'iteratore di ritorno in una
// variabile allocata e assegnando iterator_ptr all'indirizzo della 
// variabile.
// Questa soluzione che implica la duplicazione delle informazioni contenute 
// nella struttura iteratore e' solo temporanea ed e' stata operata 
// esclusivamente per comodita'. 
// Una soluzione alternativa che non comporta duplicazione prevede :
//
// - Copiare il contenuto di list.h in list_2.h.
// - Realizzare una funzione di accesso che ritorni il tipo "reference ad
//   iteratore".
// - Assegnare ad iterator_ptr l'indirizzo del reference ritornato dalla
//   funzione descritta al punto precedente (che consiste, in pratica, nel
//   modificare le funzioni push() e append() definite in list_2.h).
// - Eliminare l'operazione di "delete" associata ai distruttori definiti per
//   le strutture ausiliarie (in pratica eliminare i distruttori). 
//
// -------------------------------------------
// FUNZIONI DI INTERFACCIA AI TIPI AUSILIARI :
// -------------------------------------------
// 
// Diamo ora di seguito l'elenco delle funzioni proprie associate alle classi
// definite ricordando che i tipi v_type, e_type ed f_type sono delle 
// ridefinizioni dei tipi vertex_2< ... >, edge_2< ... > e face_2< ... > e che
// i tipi vi_type, ei_type, fi_type sono ridefinizioni dei tipi
// list_2<v_type *>::iterator, list_2<e_type *>::iterator e 
// list_2<f_type *>::iterator.
//
// -----------------------------
// Classe vertex_2_info< ... > :
// -----------------------------
//
// Funzioni di accesso "public" :
// ------------------------------
//
// - void * get_iterator_ptr() :
//   Ritorna il valore del campo iterator_ptr.
//
// - bool is_marked() :
//   Ritorna "true" se il vertice e' marcato.
//
// - f_type * containing_face() :
//   Ritorna il puntatore alla faccia associata alla relazione F-Vertex-Face.
//
// - bool is_on_some_boundary() :
//   Ritorna se il vertice appartiene al contorno di una qualche faccia. Il
//   suo corretto funzionamento dipende dall'assunzione che un vertice e'
//   un vertice feature sse non appartiene al contorno di nessuna faccia e
//   che in tal caso la relazione F-Vertex-Face ne indichi la faccia in cui
//   e' contenuto. A tal proposito, il campo rappresentante tale relazione
//   deve essere assegnato alla giusta faccia anche se il vertice si trova
//   esclusivamente su uno o piu' spigoli "feature".
//
// Funzioni di assegnazione "public" :
// -----------------------------------
//
// - void assign_iterator_ptr(void * i_p) :
//   Assegna il campo iterator_pointer al valore di i_p. 
//
// - void free_iterator_ptr() :
//   Dealloca l'area di memoria associata al valore di iterator_ptr.
//
// - void mark() :
//   Marca il vertice.
//
// - void unmark() :
//   Smarca il vertice.
//
// - void assign_containing_face(f_type * f) :
//   Assegna il punatore associato alla relazione F-Vertex_Face.
//
//
// ---------------------------
// Classe edge_2_info< ... > :
// ---------------------------
//
// Funzioni di accesso "public" :
// ------------------------------
//
// - void * get_iterator_ptr() :
//   Ritorna il valore del campo iterator_ptr.
//
// - bool is_marked() :
//   Ritorna "true" se il vertice e' marcato.
//
// Funzioni di assegnazione "public" :
// -----------------------------------
//
// - void assign_iterator_ptr(void * i_p) :
//   Assegna il campo iterator_pointer al valore di i_p. 
//
// - void free_iterator_ptr() :
//   Dealloca l'area di memoria associata al valore di iterator_ptr.
//
// - void mark() :
//   Marca lo spigolo.
//
// - void unmark() :
//   Smarca lo spigolo.
//
//
// ---------------------------
// Classe face_2_info< ... > :
// ---------------------------
//
// Funzioni di accesso "public" :
// ------------------------------
//
// - void * get_iterator_ptr() :
//   Ritorna il valore del campo iterator_ptr.
//
// - bool is_marked() :
//   Ritorna "true" se la faccia e' marcata.
//
// - vi_type f_vertices_begin() :
//   Ritorna l'iteratore al primo elemento della lista associata alla relazione
//   F-Face-Vertex.
//
// - vi_type f_vertices_end() :
//   Ritorna l'iteratore che rappresenta il past-end della lista associata alla
//   relazione F-Face-Vertex.
//
// - bool is_f_vertex(v_type * v) :
//   Ritorna "true" se v appartiene ai vertici associati alla relazione
//   F-Face-Vertex.
//
// - bool is_f_vertex(v_type * v, vi_type & vi) :
//   Ritorna "true" se v e' appartiene ai vertici associati alla relazione
//   F-Face-Vertex e, nel caso, assegna vi all'iteratore che lo rappresenta
//   nella corrispondente lista.
//
// - bool no_f_vertices() :
//   Ritorna "true" se non vi sono vertici associati alla relazione
//   F-Face-Vertex.
//
// - bool is_f_inside_edge(e_type * e) :
//   Ritorna "true" se e e' uno spigolo associato alla relazione F-Face-Edge.
//
// - bool is_f_inside_edge(e_type * e, ei_type & ei)
//   Ritorna "true" se e e' uno spigolo associato alla relazione F-Face-Edge
//   e, nel caso, assegna ei all'iteratore che lo rappresenta nella
//   corrispondente lista.
//  
// - ei_type f_inside_edges_begin() :
//   Ritorna l'iteratore al primo elemento della lista associata alla relazione
//   F-Face-Edge.
//
// - ei_type f_inside_edges_end() :
//   Ritorna l'iteratore che rappresenta il past-end della lista associata alla
//   relazione F-Face-Edge.
//
// - bool no_f_inside_edges() :
//   Ritorna "true" se non vi sono spigoli associati alla relazione
//   F-Face-Edge.
//  
// - bool no_inside_edges() :
//   Ritorna se non vi sono spigoli che rappresentano cicli interni (buchi).
//
// - ei_type inside_edges_begin() :
//   Ritorna l'iteratore al primo elemento della lista associata agli spigoli
//   che rappresentano cicli interni (buchi).
//
// - ei_type inside_edges_end() :
//   Ritorna l'iteratore che rappresenta il past-end della lista associata agli
//   spigoli che rappresentano cicli interni (buchi).
//
// - bool is_inside_edge(e_type * e) :
//   Ritorna "true" se e appartiene agli spigoli che rappresentano cicli interni
//   (buchi).
//
// - bool is_inside_edge(e_type * e, ei_type & ei) :
//   Ritorna "true" se e e' uno spigolo che rappresenta cicli interni (buci) e,
//   nel caso, assegna ei all'iteratore ad esso associato nella corrispondente
//   lista.
//
// Funzioni di assegnazione "public" :
// -----------------------------------
//
// - void assign_iterator_ptr(void * i_p) :
//   Assegna il campo iterator_pointer al valore di i_p. 
//
// - void free_iterator_ptr() :
//   Dealloca l'area di memoria associata al valore di iterator_ptr.
// 
// - void mark() :
//   Marca la faccia.
//
// - void unmark() :
//   Smarca la faccia.
//
// - void add_f_vertex(v_type * v) :
//   Inserisce v tra i vertici associati alla relazione F-Face-Vertex.
//
// - void remove_f_vertex(vi_type pos)
//   Elimina dai vertici associati alla relazione Face-Vertex quello
//   rappresentato dall'iteratore pos.
//
// - bool remove_f_vertex(v_type * v) :
//   Elimina v dai vertici associati alla relazione F-Face-Vertex. Nel caso
//   esso non appartenga a tali vertici ritorna "false".
//
// - void add_f_inside_edge(e_type * e) :
//   Inserisce e tra gli spigoli associati alla relazione F-Face-Edge.
//
// - void remove_f_inside_edge(ei_type pos) :
//   Elimina dagli spigoli associati alla relazione F-Face-Edge quello
//   rappresentato dall'iteratore pos.
//
// - bool remove_f_inside_edge(e_type * e) :
//   Elimina e dagli spigoli associati alla relazione F-Face-Edge. Nel caso
//   esso non appartenga a tali spigoli ritorna "false".
//
// - void add_inside_edge(e_type * e) :
//   Inserisce e tra gli spigoli che rappresentano cicli interni (buchi)
//
// - void remove_inside_edge(ei_type pos) :
//   Elimina dagli spigoli che rappresentano cicli interni (buchi) quello 
//   associato all'iteratore pos.
//  
// - bool remove_inside_edge(e_type * e) :
//   Elimina e dagli spigoli che rappresentano cicli interni (buchi). Nel caso
//   esso non appartenga a tali spigoli ritorna "false".
//
// ****************************************************************************

#ifndef __PEG2_INFO
#define __PEG2_INFO

#include <CGAL/Point_2.h>
#include <list_2.h>

template <class VInfo, class EInfo, class FInfo>
class vertex_2;
template <class VInfo, class EInfo, class FInfo>
class edge_2;
template <class VInfo, class EInfo, class FInfo>
class face_2;

template <class RepType, class Ext>
class face_2_info;

template <class Ext>
class edge_2_info : public Ext
{
  bool marked;
  void *  iterator_ptr;

  public:

  edge_2_info()
  { iterator_ptr = NULL; marked = false; }

  void assign_iterator_ptr(void * i_p)
  { iterator_ptr = i_p; }

  void * get_iterator_ptr()
  { return iterator_ptr; }

  void free_iterator_ptr()
  { delete iterator_ptr; }
  
  bool is_marked()
  { return marked; }
  
  void mark()
  { marked = true; }
  
  void unmark()
  { marked = false; }
};


template <class RepType, class Ext>
class vertex_2_info : public CGAL_Point_2<RepType>, public Ext
{
  typedef vertex_2_info<RepType,Ext> v_info;
  typedef edge_2_info<Ext> e_info;
  typedef face_2_info<RepType,Ext> f_info;
  typedef face_2<v_info,e_info,f_info> f_type;

  bool marked;
  void * iterator_ptr;
  f_type * VF;
  
  public:

  vertex_2_info()
  { iterator_ptr = NULL; VF = NULL; marked = false; }

  void assign_iterator_ptr(void * i_p)
  { iterator_ptr = i_p; }

  void * get_iterator_ptr()
  { return iterator_ptr; }

  void free_iterator_ptr()
  { delete iterator_ptr; }
  
  bool is_marked()
  { return marked; }
  
  void mark()
  { marked = true; }
  
  void unmark()
  { marked = false; }
  
  f_type * containing_face()
  { return VF; }

  void assign_containing_face(f_type * f)
  { VF = f; }
  
  bool is_on_some_boundary()
  { return VF == NULL; }
};

template <class RepType, class Ext>
class face_2_info : public Ext
{
  typedef vertex_2_info<RepType,Ext> v_info;
  typedef edge_2_info<Ext> e_info;
  typedef face_2_info<RepType,Ext> f_info;
  typedef vertex_2<v_info,e_info,f_info> v_type;
  typedef edge_2<v_info,e_info,f_info> e_type;
  typedef face_2<v_info,e_info,f_info> f_type;
  typedef list_2<v_type *>::iterator vi_type;
  typedef list_2<e_type *>::iterator ei_type;
  typedef list_2<f_type *>::iterator fi_type;

  bool marked;
  void * iterator_ptr;

  list_2<v_type *> FV;		// Vertici feature
  list_2<e_type *> FIE;		// Rappresentanti di spigoli feature
  list_2<e_type *> IE;		// Rappresentanti di cicli interni (buchi) 

  public:

  face_2_info()
  { iterator_ptr = NULL; marked = false; }

  ~face_2_info()
  { FV.free(); FIE.free(); IE.free(); }

  void assign_iterator_ptr(void * i_p)
  { iterator_ptr = i_p; }

  void * get_iterator_ptr()
  { return iterator_ptr; }

  void free_iterator_ptr()
  { delete iterator_ptr; }

  bool is_marked()
  { return marked; }
  
  void mark()
  { marked = true; }
  
  void unmark()
  { marked = false; }
  
  vi_type f_vertices_begin()
  { return FV.begin(); }

  vi_type f_vertices_end()
  { return FV.end(); }

  bool is_f_vertex(v_type * v)
  { return FV.is_in_list(v); }
  
  bool is_f_vertex(v_type * v, vi_type & vi)
  { return FV.is_in_list(v, vi); }
  
  void add_f_vertex(v_type * v)
  { FV.push_front(v); }

  void remove_f_vertex(vi_type pos)
  { FV.erase(pos); }

  bool remove_f_vertex(v_type * v)
  { return FV.remove(v); }

  bool no_f_vertices()
  { return FV.empty(); }

  bool is_f_inside_edge(e_type * e)
  { return FIE.is_in_list(e); }

  bool is_f_inside_edge(e_type * e, ei_type & ei)
  { return FIE.is_in_list(e, ei); }
  
  ei_type f_inside_edges_begin()
  { return FIE.begin(); }

  ei_type f_inside_edges_end()
  { return FIE.end(); }

  void add_f_inside_edge(e_type * e)
  { FIE.push_front(e); }

  void remove_f_inside_edge(ei_type pos)
  { FIE.erase(pos); }

  bool remove_f_inside_edge(e_type * e)
  { return FIE.remove(e); }
  
  bool no_f_inside_edges()
  { return FIE.empty(); }
  
  ei_type inside_edges_begin()
  { return IE.begin(); }

  ei_type inside_edges_end()
  { return IE.end(); }

  void add_inside_edge(e_type * e)
  { IE.push_front(e); }

  void remove_inside_edge(ei_type pos)
  { IE.erase(pos); }
  
  bool remove_inside_edge(e_type * e)
  { return IE.remove(e); }
   
  bool no_inside_edges()
  { return IE.empty(); }
  
  bool is_inside_edge(e_type * e)
  { return IE.is_in_list(e); } 
  
  bool is_inside_edge(e_type * e, ei_type & ei)
  { return IE.is_in_list(e, ei); } 
  
  // SOLO PER DEBUG
  void print_f_vertex_list()
  { FV.print(); }
  
  void print_f_inside_edge_list()
  { FIE.print(); }
  
  void print_inside_edge_list()
  { IE.print(); }
};

#endif
